@@ -16,8 +16,8 @@
 
   You should have received a copy of the GNU Lesser General Public License
   along with the ECM Library; see the file COPYING.LIB.  If not, write to
-  the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
-  MA 02111-1307, USA.
+  the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
+  MA 02110-1301, USA.
 */
 
 #include <stdlib.h>
@@ -30,18 +30,16 @@
 #endif
 
 #if (MULT == KS)
-#define LIST_MULT_N kronecker_schonhage
-#ifdef HAVE___GMPN_MUL_FFT
-#define WRAP /* use wrap-around multiplication for low short product */
-#endif
+  #define LIST_MULT_N kronecker_schonhage
+  #define WRAP /* use wrap-around multiplication for low short product */
 #elif (MULT == TOOM4)
-#define LIST_MULT_N toomcook4
+  #define LIST_MULT_N toomcook4
 #elif (MULT == TOOM3)
-#define LIST_MULT_N toomcook3
+  #define LIST_MULT_N toomcook3
 #elif (MULT == KARA)
-#define LIST_MULT_N karatsuba
+  #define LIST_MULT_N karatsuba
 #else
-#error "MULT is neither KS, TOOM4, nor TOOM3, nor KARA"
+  #error "MULT is neither KS, TOOM4, nor TOOM3, nor KARA"
 #endif
 
 extern unsigned int Fermat;
@@ -80,7 +78,7 @@ init_list (unsigned int n)
 }
 
 /* creates a list of n integers, return NULL if error. Allocates each
-   mpz_t to the size of n */
+   mpz_t to the size of N bits */
 listz_t
 init_list2 (unsigned int n, unsigned int N)
 {
@@ -180,14 +178,14 @@ list_set (listz_t p, listz_t q, unsigned int n)
     mpz_set (p[i], q[i]);
 }
 
-/* p[0] <-> p[n], p[1] <-> p[n-1], ... */
+/* p[0] <-> p[n-1], p[1] <-> p[n-2], ... */
 void
 list_revert (listz_t p, unsigned int n)
 {
   unsigned int i;
 
-  for (i = 0; i < n - i; i++)
-    mpz_swap (p[i], p[n - i]);
+  for (i = 0; i < n - 1 - i; i++)
+    mpz_swap (p[i], p[n - 1 - i]);
 }
 
 void
@@ -280,7 +278,7 @@ list_gcd (mpz_t p, listz_t l, unsigned int k, mpz_t n)
    product (mod n) of itself and all previous entries */
    
 void 
-list_mulup (mpz_t p, listz_t l, unsigned int k, mpz_t n, mpz_t t)
+list_mulup (listz_t l, unsigned int k, mpz_t n, mpz_t t)
 {
   unsigned int i;
   
@@ -312,6 +310,7 @@ list_mul_low (listz_t a, listz_t b, listz_t c, unsigned int K, listz_t t,
 {
   unsigned int p, q;
 
+  ASSERT(K > 0);
   switch (K)
     {
     case 1:
@@ -356,6 +355,7 @@ list_mul_high (listz_t a, listz_t b, listz_t c, unsigned int K, listz_t t)
 #else
   unsigned int p, q;
 
+  ASSERT(K > 0);
   switch (K)
     {
     case 1:
@@ -525,7 +525,7 @@ list_mul (listz_t a, listz_t b, unsigned int k, int monic_b,
 {
   unsigned int i, po2;
 
-  ASSERTD (k >= l);
+  ASSERT(k == l || k == l + 1);
   
   for (po2 = l; (po2 & 1) == 0; po2 >>= 1);
   po2 = (po2 == 1);
@@ -607,40 +607,81 @@ list_mulmod (listz_t a2, listz_t a, listz_t b, listz_t c, unsigned int k,
 /* puts in G[0]..G[k-1] the coefficients from (x-a[0])...(x-a[k-1])
    Warning: doesn't fill the coefficient 1 of G[k], which is implicit.
    Needs k + list_mul_mem(k/2) cells in T.
+   G == a is permissible. T must not overlap with anything else
 */
 void
 PolyFromRoots (listz_t G, listz_t a, unsigned int k, listz_t T, mpz_t n)
 {
   unsigned int l, m;
 
-   if (k <= 1)
-     {
-       mpz_mod (G[0], a[0], n);
-       return;
-     }
+  ASSERT (T != G && T != a);
 
-    /* (x-a[0]) * (x-a[1]) = x^2 - (a[0]+a[1]) * x + a[0]*a[1]
-       however we construct (x+a[0]) * (x+a[1]) instead, i.e. the
-       polynomial with the opposite roots. This has no consequence if
-       we do it for all polynomials: if F(x) and G(x) have a common root,
-       then so do F(-x) and G(-x). This saves one negation.
-     */
-   if (k == 2)
-     {
-       mpz_mul (T[0], a[0], a[1]);
-       mpz_add (T[1], a[1], a[0]); /* mpz_add may allocate extra limb */
-       mpz_mod (G[1], T[1], n);
-       mpz_mod (G[0], T[0], n);
-       return;
-     }
+  if (k <= 1)
+    {
+      if (k == 1)
+	{
+#if NEGATED_ROOTS == 1
+	  mpz_mod (G[0], a[0], n);
+#else
+	  mpz_neg (T[0], a[0]);
+	  mpz_mod (G[0], T[0], n);
+#endif
+	}
+      return;
+    }
 
-   m = k / 2;
-   l = k - m;
-
-   PolyFromRoots (G, a, l, T, n);
-   PolyFromRoots (G + l, a + l, m, T, n);
-   list_mul (T, G, l, 1, G + l, m, 1, T + k);
-   list_mod (G, T, k, n);
+#if 0
+  /* (x-a[0]) * (x-a[1]) = x^2 - (a[0]+a[1]) * x + a[0]*a[1]
+     however we construct (x+a[0]) * (x+a[1]) instead, i.e. the
+     polynomial with the opposite roots. This has no consequence if
+     we do it for all polynomials: if F(x) and G(x) have a common root,
+     then so do F(-x) and G(-x). This saves one negation.
+     
+     NOT ANY MORE! We added the mpz_neg()'s  so building the poly from
+     its roots works as one would expect in the FastPM1Stage2 algorithm!
+  */
+  if (k == 2)
+    {
+      mpz_mul (T[0], a[0], a[1]);
+      mpz_add (T[1], a[1], a[0]); /* mpz_add may allocate extra limb */
+#if NEGATED_ROOTS == 0
+      mpz_neg (T[1], T[1]);
+#endif
+      mpz_mod (G[1], T[1], n);
+      mpz_mod (G[0], T[0], n);
+      return;
+    }
+#endif
+  
+  m = k / 2; /* m == 1 */
+  l = k - m; /* l == 2 */
+  
+  PolyFromRoots (G, a, l, T, n);
+  PolyFromRoots (G + l, a + l, m, T, n);
+#if 0
+  {
+    unsigned int i;
+    outputf (OUTPUT_RESVERBOSE, "N=%Zd;", n);
+    outputf (OUTPUT_RESVERBOSE, "(x^%u ", l);
+    for (i = 0; i < l; i++)
+      outputf (OUTPUT_RESVERBOSE, "+ (Mod(%Zd, N) * x^%u) ", G[i], i);
+    outputf (OUTPUT_RESVERBOSE, ") * ( x^%u", m);
+    for (i = 0; i < m; i++)
+      outputf (OUTPUT_RESVERBOSE, "+ (Mod(%Zd, N) * x^%u) ", G[i + l], i);
+    outputf (OUTPUT_RESVERBOSE, ")  ");
+  }
+#endif
+  list_mul (T, G, l, 1, G + l, m, 1, T + k);
+  list_mod (G, T, k, n);
+#if 0
+  {
+    unsigned int i;
+    outputf (OUTPUT_RESVERBOSE, " == ( x^%u ", k);
+    for (i = 0; i < k; i++)
+      outputf (OUTPUT_RESVERBOSE, "+ (Mod(%Zd, N) * x^%u) ", G[i], i);
+    outputf (OUTPUT_RESVERBOSE, ")\n");
+  } 
+#endif
 }
 
 /* puts in G[0]..G[k-1] the coefficients from (x-a[0])...(x-a[k-1])
@@ -667,7 +708,15 @@ PolyFromRoots_Tree (listz_t G, listz_t a, unsigned int k, listz_t T,
   if (k <= 1)
     {
       if (k == 1)
-        mpz_set (G[0], a[0]);
+        {
+#if NEGATED_ROOTS == 1
+	  mpz_mod (G[0], a[0], n);
+#else
+	  mpz_neg (a[0], a[0]);
+	  mpz_mod (G[0], a[0], n);
+          mpz_neg (a[0], a[0]);
+#endif
+        }
       return 0;
     }
 
@@ -688,6 +737,9 @@ PolyFromRoots_Tree (listz_t G, listz_t a, unsigned int k, listz_t T,
      polynomial with the opposite roots. This has no consequence if
      we do it for all polynomials: if F(x) and G(x) have a common root,
      then so do F(-x) and G(-x). This saves one negation.
+       
+       NOT ANY MORE! We added the mpz_neg() below so building the poly from
+       its roots works as one would expect in the FastPM1Stage2 algorithm!
   */
   if (k == 2)
     {
@@ -695,6 +747,9 @@ PolyFromRoots_Tree (listz_t G, listz_t a, unsigned int k, listz_t T,
       mpz_set (H1[1], a[1]);
       mpz_mul (T[0], a[0], a[1]);
       mpz_add (G[1], a[1], a[0]);
+#if NEGATED_ROOTS == 0
+      mpz_neg (G[1], G[1]);
+#endif
       mpz_mod (G[1], G[1], n);
       mpz_mod (G[0], T[0], n);
       return 0;
@@ -768,10 +823,10 @@ PolyInvert (listz_t q, listz_t b, unsigned int K, listz_t t, mpz_t n)
         }
       else if (po2)
         {
-          list_revert (q + k, l - 1);
+          list_revert (q + k, l);
           /* This expects the leading monomials explicitly in q[2k-1] and b[k+l-1] */
-          F_mul_trans (t, q + k, b, K, Fermat, t + k);
-          list_revert (q + k, l - 1);
+          F_mul_trans (t, q + k, b, K / 2, K, Fermat, t + k);
+          list_revert (q + k, l);
           list_neg (t, t, k, n);
         }
       else
