@@ -14,50 +14,61 @@
 
   You should have received a copy of the GNU General Public License along
   with this program; see the file COPYING.  If not, write to the Free
-  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
+  Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
   02111-1307, USA.
 */
 
-#include <ctype.h>
-#include <math.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <time.h>
-#if !defined (_MSC_VER) && !defined (__MINGW32__)
-#include <unistd.h>
-#include <sys/time.h>
-#include <sys/resource.h> /* for setpriority */
-#else
-#include <io.h>		/* for access() */
-#define F_OK 0
+#include <math.h>
+#ifdef _MSC_VER
+#  include <winsock2.h>
 #endif
-#include "gmp.h"
-#include "ecm.h"
 #include "ecm-ecm.h"
+
+#ifdef HAVE_UNISTD_H /* for access() */
+# include <unistd.h>
+#else
+# define F_OK 0
+# ifdef HAVE_IO_H
+#  include <io.h>
+# endif
+#endif
+
+#ifdef HAVE_SIGNAL_H
+# include <signal.h>
+#endif
+
 #ifdef HAVE_GWNUM
 /* For GWNUM_VERSION */
 #include "gwnum.h"
 #endif
 
-#if defined (__MINGW32__) || defined (_MSC_VER) /* || defined (__CYGWIN__) */
-/* needed for priority setting */               /* not sure about CyGwin and windows.h and Win32 API functions */
-#include <windows.h>
-#endif
-
 /* #define DEBUG */
 
 /* people keeping track of champions and corresponding url's: ECM, P-1, P+1 */
-char *champion_keeper[3] =
-{ "Richard Brent <rpb@comlab.ox.ac.uk>",
+static char *champion_keeper[3] =
+{ "Richard Brent <champs@rpbrent.com>",
   "Paul Zimmermann <zimmerma@loria.fr>",
   "Paul Zimmermann <zimmerma@loria.fr>"};
-char *champion_url[3] =
-{"ftp://ftp.comlab.ox.ac.uk/pub/Documents/techpapers/Richard.Brent/champs.txt",
+static char *champion_url[3] =
+{"http://wwwmaths.anu.edu.au/~brent/ftp/champs.txt",
  "http://www.loria.fr/~zimmerma/records/Pminus1.html",
  "http://www.loria.fr/~zimmerma/records/Pplus1.html"};
 /* minimal number of digits to enter the champions table for ECM, P-1, P+1 */
-unsigned int champion_digits[3] = { 53, 43, 37 };
+static unsigned int champion_digits[3] = { 61, 50, 40 };
+
+/* probab_prime_p() can get called from other modules. Instead of passing
+   prpcmd to those functions, we make it static here - this variable will
+   be set only in main, and read only in probab_prime_p() */
+#ifdef WANT_SHELLCMD
+static  char *prpcmd = NULL;
+#endif
+
+static int exit_asap_value = 0;
+static int exit_asap_signalnr = 0; /* Remembers which signal we received */
 
 /* Tries to read a number from a line from fd and stores it in r.
    Keeps reading lines until a number is found. Lines beginning with "#"
@@ -118,7 +129,60 @@ new_line:
   return 1;
 }
 
-void usage (void)
+int 
+probab_prime_p (mpz_t N, int reps)
+{
+#ifdef WANT_SHELLCMD
+  if (prpcmd != NULL)
+    {
+      FILE *fc;
+      int r;
+      fc = popen (prpcmd, "w");
+      if (fc != NULL)
+        {
+          gmp_fprintf (fc, "%Zd\n", N);
+          r = pclose (fc);
+          if (r == 0) /* Exit status of 0 means success = is a PRP */
+            return 1;
+          else
+            return 0;
+        } else {
+          fprintf (stderr, "Error executing the PRP command\n");
+          exit (EXIT_FAILURE);
+        }
+    } else
+#endif
+      return mpz_probab_prime_p (N, reps);
+}
+
+void 
+signal_handler (int sig)
+{
+  if (sig == SIGINT || sig == SIGTERM)
+    {
+      exit_asap_value = 1;
+      exit_asap_signalnr = sig;
+      /* If one of these two signals arrives again, we'll let the default 
+         handler take over,  which will usually terminate the process 
+         immediately. */
+      signal (SIGINT, SIG_DFL);
+      signal (SIGTERM, SIG_DFL);
+    }
+  else
+    {
+      /* How did this happen? Let's ignore it for now, abort instead? */
+    }
+}
+
+int
+stop_asap_test ()
+{
+  return exit_asap_value;
+}
+
+
+static void 
+usage (void)
 {
     printf ("Usage: ecm [options] B1 [[B2min-]B2] < file\n");
     printf ("\nParameters:\n");
@@ -137,16 +201,27 @@ void usage (void)
     printf ("  -q           quiet mode\n");
     printf ("  -v           verbose mode\n");
     printf ("  -timestamp   print a time stamp with each number\n");
-    printf ("  -mpzmod      use GMP's mpz_mod for mod reduction\n");
-    printf ("  -modmuln     use Montgomery's MODMULN for mod reduction\n");
-    printf ("  -redc        use Montgomery's REDC for mod reduction\n");
+    printf ("  -mpzmod      use GMP's mpz_mod for modular reduction\n");
+    printf ("  -modmuln     use Montgomery's MODMULN for modular reduction\n");
+    printf ("  -redc        use Montgomery's REDC for modular reduction\n");
     printf ("  -nobase2     disable special base-2 code\n");
-    printf ("  -base2 n     force base 2 mode with 2^n+1 (n>0) or 2^n-1 (n<0)\n");
+    printf ("  -base2 n     force base 2 mode with 2^n+1 (n>0) or 2^|n|-1 (n<0)\n");
+    printf ("  -ntt         enable NTT convolution routines in stage 2\n");
+    printf ("  -no-ntt      disable NTT convolution routines in stage 2\n");
     printf ("  -save file   save residues at end of stage 1 to file\n");
     printf ("  -savea file  like -save, appends to existing files\n");
     printf ("  -resume file resume residues from file, reads from stdin if file is \"-\"\n");
+    printf ("  -chkpnt file save periodic checkpoints during stage 1 to file\n");
     printf ("  -primetest   perform a primality test on input\n");
     printf ("  -treefile f  store product tree of F in files f.0 f.1 ... \n");
+    printf ("  -maxmem n    use at most n MB of memory in stage 2\n");
+    printf ("  -stage1time n add n seconds to ECM stage 1 time (for expected time est.)\n");
+#ifdef WANT_SHELLCMD
+    printf ("  -faccmd cmd  execute cmd when factor is found. Input number, factor\n"
+            "               and cofactor are given to cmd via stdin, each on a line\n");
+    printf ("  -prpcmd cmd  use shell command cmd to do prp tests (number via stdin)\n");
+    printf ("  -idlecmd cmd before each curve run cmd and terminate if exit code >0\n");
+#endif
 
     /*printf ("  -extra functions added by JimF\n"); */
     printf ("  -i n         increment B1 by this constant on each run\n");
@@ -164,14 +239,7 @@ void usage (void)
     printf ("  -go val      Preload with group order val, which can be a simple expression,\n");
     printf ("               or can use N as a placeholder for the number being factored.\n");
 
-    /*printf ("  -extra functions added by PhilC\n"); */
-    printf ("  -prp cmd     use shell command cmd to do large primality tests\n");
-    printf ("  -prplen n    only candidates longer than this number of digits are 'large'\n");
-    printf ("  -prpval n    value>=0 which indicates the prp command foundnumber to be PRP.\n");
-    printf ("  -prptmp file outputs n value to temp file prior to running (NB. gets deleted)\n");
-    printf ("  -prplog file otherwise get PRP results from this file (NB. gets deleted)\n");
-    printf ("  -prpyes str  literal string found in prplog file when number is PRP\n");
-    printf ("  -prpno str   literal string found in prplog file when number is composite\n");
+
     printf ("  -h, --help   Prints this help and exit.\n");
 }
 
@@ -186,23 +254,25 @@ void usage (void)
 int
 main (int argc, char *argv[])
 {
+  char **argv0 = argv;
   mpz_t x, sigma, A, f, orig_x0, B2, B2min, startingB2min;
   mpcandi_t n;
   mpgocandi_t go;
   mpq_t rat_x0;
   double B1, B1done;
-  int result = 0;
+  int result = 0, returncode = 0;
   int verbose = OUTPUT_NORMAL; /* verbose level */
   int timestamp = 0;
   int method = ECM_ECM, method1;
+  int use_ntt = 1;     /* Default, use NTT if input is small enough */
   int specific_x0 = 0, /* 1=starting point supplied by user, 0=random or */
                        /* compute from sigma */
       specific_sigma = 0;  /* 1=sigma from command line, 0=make random */
   int factor_is_prime;
         /* If a factor was found, indicate whether factor, cofactor are */
         /* prime. If no factor was found, both are zero. */
-  int repr = ECM_DEFAULT_REPR; /* automatic choice */
-  int k = ECM_DEFAULT_K; /* default number of blocks in stage 2 */
+  int repr = ECM_MOD_DEFAULT; /* automatic choice */
+  unsigned long k = ECM_DEFAULT_K; /* default number of blocks in stage 2 */
   int S = ECM_DEFAULT_S;
              /* Degree for Brent-Suyama extension requested by user.
                 Positive value: use S-th power,
@@ -210,9 +280,12 @@ main (int argc, char *argv[])
                 default (0): automatic choice. */
   gmp_randstate_t randstate;
   char *savefilename = NULL, *resumefilename = NULL, *infilename = NULL;
-  char *TreeFilename = NULL;
+  char *TreeFilename = NULL, *chkfilename = NULL;
   char rtime[256] = "", who[256] = "", comment[256] = "", program[256] = "";
   FILE *savefile = NULL, *resumefile = NULL, *infile = NULL;
+  mpz_t resume_lastN, resume_lastfac; /* When resuming residues from a file,
+        store the last number processed and the factors found for this it */
+  int resume_wasPrp = 0; /* 1 if resume_lastN/resume_lastfac is a PRP */
   int primetest = 0, saveappend = 0;
   double autoincrementB1 = 0.0, startingB1;
   unsigned int autoincrementB1_calc = 0;
@@ -228,7 +301,13 @@ main (int argc, char *argv[])
   unsigned int decimal_cofactor = 0;
   double maxtrialdiv = 0.0;
   double B2scale = 1.0;
+  double maxmem = 0.;
+  double stage1time = 0.;
   ecm_params params;
+#ifdef WANT_SHELLCMD
+  char *faccmd = NULL;
+  char *idlecmd = NULL;
+#endif
 
   /* check ecm is linked with a compatible library */
   if (mp_bits_per_limb != GMP_NUMB_BITS)
@@ -248,11 +327,11 @@ main (int argc, char *argv[])
   mpgocandi_t_init (&go);
 
   /* Init variables we might need to store options */
-  mpz_init (sigma);
-  mpz_init (A);
-  mpz_init (B2);
-  mpz_init (B2min);
-  mpz_init (startingB2min);
+  MPZ_INIT (sigma);
+  MPZ_INIT (A);
+  MPZ_INIT (B2);
+  MPZ_INIT (B2min);
+  MPZ_INIT (startingB2min);
   mpq_init (rat_x0);
 
   /* first look for options */
@@ -290,28 +369,40 @@ main (int argc, char *argv[])
 	}
       else if (strcmp (argv[1], "-mpzmod") == 0)
         {
-          repr = 1;
+          repr = ECM_MOD_MPZ;
 	  argv++;
 	  argc--;
         }
       else if (strcmp (argv[1], "-modmuln") == 0)
         {
-          repr = 2;
+          repr = ECM_MOD_MODMULN;
 	  argv++;
 	  argc--;
         }
       else if (strcmp (argv[1], "-redc") == 0)
         {
-          repr = 3;
+          repr = ECM_MOD_REDC;
 	  argv++;
 	  argc--;
         }
       else if (strcmp (argv[1], "-nobase2") == 0)
         {
-          repr = -1;
+          repr = ECM_MOD_NOBASE2;
 	  argv++;
 	  argc--;
         }
+      else if (strcmp (argv[1], "-ntt") == 0)
+	{
+	  use_ntt = 2; /* Use NTT, even for large input numbers */
+	  argv++;
+	  argc--;
+	}
+      else if (strcmp (argv[1], "-no-ntt") == 0)
+	{
+	  use_ntt = 0; /* Never use NTT */
+	  argv++;
+	  argc--;
+	}
       else if (strcmp (argv[1], "-primetest") == 0)
         {
           primetest = 1;
@@ -349,52 +440,18 @@ main (int argc, char *argv[])
 	  argv++;
 	  argc--;
         }
-#if defined (__MINGW32__) || (_MSC_VER)
       else if (strcmp (argv[1], "-n") == 0)
         {
-/*	  fprintf (stderr, "Executing at lower priority\n"); */
-	  SetPriorityClass (GetCurrentProcess (), IDLE_PRIORITY_CLASS);
-	  SetThreadPriority (GetCurrentThread (), THREAD_PRIORITY_BELOW_NORMAL);
+          NICE10;
 	  argv++;
 	  argc--;
         }
       else if (strcmp (argv[1], "-nn") == 0)
         {
-/*	  fprintf (stderr, "Executing at idle priority\n"); */
-	  SetPriorityClass (GetCurrentProcess (), IDLE_PRIORITY_CLASS);
-	  SetThreadPriority (GetCurrentThread (), THREAD_PRIORITY_IDLE);
+          NICE20;
 	  argv++;
 	  argc--;
         }
-#elif __CYGWIN__
-      else if (strcmp (argv[1], "-n") == 0)
-        {
-	  /* Increase niceness by 10 */
-	  nice (10);
-	  argv++;
-	  argc--;
-        }
-      else if (strcmp (argv[1], "-nn") == 0)
-        {
-	  /* Increase niceness by 20 */
-	  nice (20);
-	  argv++;
-	  argc--;
-        }
-#else
-      else if (strcmp (argv[1], "-n") == 0)
-        {
-          setpriority (PRIO_PROCESS, 0, 10);
-	  argv++;
-	  argc--;
-        }
-      else if (strcmp (argv[1], "-nn") == 0)
-        {
-          setpriority (PRIO_PROCESS, 0, 20);
-	  argv++;
-	  argc--;
-        }
-#endif
       else if ((argc > 2) && (strcmp (argv[1], "-x0")) == 0)
         {
           if (mpq_set_str (rat_x0, argv[2], 0))
@@ -443,7 +500,7 @@ main (int argc, char *argv[])
         }
       else if ((argc > 2) && (strcmp (argv[1], "-k") == 0))
 	{
-	  k = atoi (argv[2]);
+	  k = atol (argv[2]);
 	  /* should this be validated? and a error/abort issued if 0 ??? */
 	  argv += 2;
 	  argc -= 2;
@@ -475,6 +532,12 @@ main (int argc, char *argv[])
 	  argv += 2;
 	  argc -= 2;
 	}
+      else if ((argc > 2) && (strcmp (argv[1], "-chkpnt") == 0))
+	{
+	  chkfilename = argv[2];
+	  argv += 2;
+	  argc -= 2;
+	}
       else if ((argc > 2) && (strcmp (argv[1], "-treefile") == 0))
 	{
 	  TreeFilename = argv[2];
@@ -557,6 +620,18 @@ main (int argc, char *argv[])
 	  argv += 2;
 	  argc -= 2;
 	}
+      else if ((argc > 2) && (strcmp (argv[1], "-maxmem") == 0))
+	{
+	  maxmem = atof (argv[2]) * 1048576.;
+	  argv += 2;
+	  argc -= 2;
+	}
+      else if ((argc > 2) && (strcmp (argv[1], "-stage1time") == 0))
+	{
+	  stage1time = atof (argv[2]);
+	  argv += 2;
+	  argc -= 2;
+	}
       else if ((argc > 2) && (strcmp (argv[1], "-go") == 0))
 	{
 	  if (go.cpOrigExpr)
@@ -580,50 +655,26 @@ main (int argc, char *argv[])
 	  argv += 2;
 	  argc -= 2;
 	}
-
-     else if ((argc > 2) && (strcmp (argv[1], "-prp") == 0))
+#ifdef WANT_SHELLCMD
+     else if ((argc > 2) && (strcmp (argv[1], "-prpcmd") == 0))
        {
-         externalprp = argv[2];
+         prpcmd = argv[2];
          argv += 2;
          argc -= 2;
        }
-     else if ((argc > 2) && (strcmp (argv[1], "-prptmp") == 0))
+     else if ((argc > 2) && (strcmp (argv[1], "-faccmd") == 0))
        {
-         externalinputprpfile = argv[2];
+         faccmd = argv[2];
          argv += 2;
          argc -= 2;
        }
-     else if ((argc > 2) && (strcmp (argv[1], "-prplog") == 0))
+     else if ((argc > 2) && (strcmp (argv[1], "-idlecmd") == 0))
        {
-         externallog = argv[2];
+         idlecmd = argv[2];
          argv += 2;
          argc -= 2;
        }
-     else if ((argc > 2) && (strcmp (argv[1], "-prpyes") == 0))
-       {
-         externalisprp = argv[2];
-         argv += 2;
-         argc -= 2;
-       }
-     else if ((argc > 2) && (strcmp (argv[1], "-prpno") == 0))
-       {
-         externaliscomposite = argv[2];
-         argv += 2;
-         argc -= 2;
-       }
-     else if ((argc > 2) && (strcmp (argv[1], "-prplen") == 0))
-       {
-         externalprplen = atoi (argv[2]);
-         argv += 2;
-         argc -= 2;
-       }
-     else if ((argc > 2) && (strcmp (argv[1], "-prpval") == 0))
-       {
-         externalprpval = atoi (argv[2]);
-         argv += 2;
-         argc -= 2;
-       }
-
+#endif
       else
 	{
 	  fprintf (stderr, "Unknown option: %s\n", argv[1]);
@@ -631,20 +682,21 @@ main (int argc, char *argv[])
 	}
     }
 
-  /* check that S is even for P-1 */
-  if ((method == ECM_PM1) && (S % 2 != 0))
+  /* check that S is even for old P-1 stage 2 */
+  if ((method == ECM_PM1) && (S != ECM_DEFAULT_S) && (S % 2 != 0))
     {
       fprintf (stderr, "Error, S should be even for P-1\n");
       exit (EXIT_FAILURE);
     }
 
-  /* Ok, now we can "reset" the breadthfirst switch so that we do depthfirst as requested */
+  /* Ok, now we can "reset" the breadthfirst switch so that we do depthfirst 
+     as requested */
   if (breadthfirst == -1)
     breadthfirst = 0;
 
   if (argc < 2)
     {
-      fprintf (stderr, "Invalid arguments. See ecm --help.\n");
+      fprintf (stderr, "Invalid arguments. See %s --help.\n", argv0[0]);
       exit (EXIT_FAILURE);
     }
 
@@ -669,12 +721,28 @@ main (int argc, char *argv[])
 	  printf ("ECM");
 	}
       printf ("]\n");
+#ifdef HAVE_GETHOSTNAME
+  if (verbose >= 2)
+    {
+#define MNAMESIZE  64
+      char mname[MNAMESIZE];
+      if (gethostname (mname, MNAMESIZE) == 0)
+        {
+          mname[MNAMESIZE - 1] = 0; /* gethostname() may omit trailing 0 */
+          printf ("Running on %s\n", mname);
+        }
+    }
+#endif
+
 #ifdef HAVE_GWNUM
-      printf ("Due to incompatible licenses, this binary file must not "
-              "be distributed.\n");
+#ifdef gwnum_is_gpl
+      if (! gwnum_is_gpl())
+#endif
+        printf ("Due to incompatible licenses, this binary file must not "
+                "be distributed.\n");
 #endif
     }
-
+  
   /* set first stage bound B1 */
   B1 = strtod (argv[1], &argv[1]);
   if (*argv[1] == '-')
@@ -719,7 +787,19 @@ main (int argc, char *argv[])
         endptr = NULL;
       
       c = -1;
-      gmp_sscanf (argv[2], "%Zd%n", B2, &c); /* Try parsing as integer */
+      {
+	int r;
+
+	r = gmp_sscanf (argv[2], "%Zd%n", B2, &c); /* Try parsing as integer */
+	if (r <= 0)
+	  {
+	    /* restore original value */
+	    if (endptr != NULL)
+	      *(--endptr) = '-';
+	    fprintf (stderr, "Invalid B2 value: %s\n", argv[2]);
+	    exit (EXIT_FAILURE);
+	  }
+      }
 #ifdef __MINGW32__
       /* MinGW scanf() returns a value 1 too high for %n */
       /* Reported to MinGW as bug number 1163607 */
@@ -774,11 +854,13 @@ main (int argc, char *argv[])
   params->verbose = verbose;
   params->method = method;
   mpz_set (params->B2, B2);
-  params->B2scale = B2scale;
   params->k = k;
   params->S = S;
   params->repr = repr;
+  params->chkfilename = chkfilename;
   params->TreeFilename = TreeFilename;
+  params->maxmem = maxmem;
+  params->stage1time = stage1time;
 
   /* Open resume file for reading, if resuming is requested */
   if (resumefilename != NULL)
@@ -794,6 +876,9 @@ main (int argc, char *argv[])
                    resumefilename);
           exit (EXIT_FAILURE);
         }
+      mpz_init (resume_lastN);
+      mpz_init (resume_lastfac);
+      mpz_set_ui (resume_lastfac, 1);
     }
 
   /* Open save file for writing, if saving is requested */
@@ -823,14 +908,29 @@ main (int argc, char *argv[])
     }
 
   mpcandi_t_init (&n); /* number(s) to factor */
-  mpz_init (f); /* factor found */
-  mpz_init (x); /* stage 1 residue */
-  mpz_init (orig_x0); /* starting point, for save file */
+  MPZ_INIT (f); /* factor found */
+  MPZ_INIT (x); /* stage 1 residue */
+  MPZ_INIT (orig_x0); /* starting point, for save file */
 
   /* We may need random numbers for sigma/starting point */
   gmp_randinit_default (randstate);
   gmp_randseed_ui (randstate, get_random_ui ());
 
+
+  /* Install signal handlers */
+#ifdef HAVE_SIGNAL
+  /* We catch signals only if there is a savefile. Otherwise there's nothing
+     we could save by exiting cleanly, but the waiting for the code to check
+     for signals may delay program end unacceptably */
+
+  if (savefile)
+    {
+      signal (SIGINT, &signal_handler);
+      signal (SIGTERM, &signal_handler);
+      params->stop_asap = &stop_asap_test;
+    }
+#endif
+
   /* loop for number in standard input or file */
 
   startingB1 = B1;
@@ -842,8 +942,8 @@ main (int argc, char *argv[])
   if (breadthfirst == 1)
     {
       breadthfirst_maxcnt = count;
-      count=1;
-      breadthfirst_cnt=0;
+      count = 1;
+      breadthfirst_cnt = 0;
     }
 
 BreadthFirstDoAgain:;
@@ -910,10 +1010,11 @@ BreadthFirstDoAgain:;
 	}
     }
 
-  while ((breadthfirst && linenum < nCandidates) || feof (infile) == 0)
+  while (((breadthfirst && linenum < nCandidates) || feof (infile) == 0)
+         && !exit_asap_value)
     {
       trial_factor_found = 0;
-      if (resumefile != NULL)
+      if (resumefile != NULL) /* resume case */
         {
 	  if (count != 1)
 	    {
@@ -924,8 +1025,29 @@ BreadthFirstDoAgain:;
           if (!read_resumefile_line (&method, x, &n, sigma, A, orig_x0, 
                 &B1done, program, who, rtime, comment, resumefile))
             break;
+          
+          if (mpz_cmp (n.n, resume_lastN) == 0)
+            {
+              /* Aha, we're trying the same number again. */
+              /* We skip this attempt if: 1. the remaining cofactor after
+                 the last attempt was a probable prime, or 2. if a factor
+                 was found and the user gave the -one option */
+              if (resume_wasPrp || 
+                  (deep == 0 && mpz_cmp_ui (resume_lastfac, 1) != 0))
+                  continue;
+              
+              /* If we found a factor in an earlier attempt, divide it out */
+              if (mpz_cmp_ui (resume_lastfac, 1) > 0)
+                mpcandi_t_addfoundfactor (&n, resume_lastfac, 1);
+            } else {
+              /* It's a different number. Set resume_lastN and 
+                 resume_lastfac */
+              mpz_set (resume_lastN, n.n);
+              mpz_set_ui (resume_lastfac, 1);
+              resume_wasPrp = n.isPrp;
+            }
 
-	  cnt = count;
+	  cnt = count; /* i.e. 1 */
 
           if (verbose >= 1)
             {
@@ -994,7 +1116,7 @@ BreadthFirstDoAgain:;
 		  exit (EXIT_FAILURE);
 		}
 
-              mpz_init (inv);
+              MPZ_INIT (inv);
               mpz_invert (inv, mpq_denref (rat_x0), n.n);
               mpz_mul (inv, mpq_numref (rat_x0), inv);
               mpz_mod (x, inv, n.n);
@@ -1028,17 +1150,10 @@ BreadthFirstDoAgain:;
 	{
 	  if ((!breadthfirst && cnt == count) || (breadthfirst && 1 == breadthfirst_cnt))
 	    {
-              if (timestamp)
-                {
-                  time_t t;
-                  
-                  t = time (NULL);
-                  printf ("[%.24s]\n", ctime (&t));
-                }
 	      /* first time this candidate has been run (if looping more than once */
-	      if (n.cpExpr && n.nexprlen < 1000)
+	      if (n.cpExpr && n.nexprlen < MAX_NUMBER_PRINT_LEN)
 		printf ("Input number is %s (%u digits)\n", n.cpExpr, n.ndigits);
-	      else if (n.ndigits < 1000)
+	      else if (n.ndigits < MAX_NUMBER_PRINT_LEN)
 		{
                   char *s;
                   s = mpz_get_str (NULL, 10, n.n);
@@ -1046,7 +1161,22 @@ BreadthFirstDoAgain:;
                   FREE (s, n.ndigits + 1);
 		}
 	      else
-		printf ("Input number has %u digits\n", n.ndigits);
+	        {
+	          /* Print only first and last ten digits of the number */
+	          mpz_t t, u;
+	          mpz_init (t);
+	          mpz_init (u);
+	          mpz_ui_pow_ui (u, 5, n.ndigits - 10);
+	          mpz_tdiv_q_2exp (t, n.n, n.ndigits - 10);
+	          mpz_tdiv_q (t, t, u);
+		  gmp_printf ("Input number is %Zd...", t);
+		  mpz_ui_pow_ui (u, 10, 10);
+		  mpz_tdiv_r (t, n.n, u);
+		  gmp_printf ("%Zd (%u digits)\n", t, n.ndigits);
+		  mpz_clear (u);
+		  mpz_clear (t);
+                }
+              
 	      if (n.isPrp)
 		printf ("****** Warning: input is probably prime ******\n");
 	    }
@@ -1068,18 +1198,22 @@ BreadthFirstDoAgain:;
 	    }
 	  fflush (stdout);
 	}
-      /* Even in verbose=0 we should primality check if told to do so, however,
-         we will print to stderr to keep stdout "clean"
+      /* Even in verbose=0 we should primality check if told to do so, 
+         however, we will print to stderr to keep stdout "clean"
          for verbose=0 like behavior */
-      else if (((!breadthfirst && cnt == count) || (breadthfirst && breadthfirst_cnt==1)) && n.isPrp)
+      else if (((!breadthfirst && cnt == count) || 
+                (breadthfirst && breadthfirst_cnt==1)) && n.isPrp)
 	{
 	  char *s;
 	  s = mpz_get_str (NULL, 10, n.n);
-	  fprintf (stderr, "Input number is %s (%u digits)\n****** Warning: input is probably prime ******\n", s, n.ndigits);
+	  fprintf (stderr, "Input number is %s (%u digits)\n"
+	           "****** Warning: input is probably prime ******\n", 
+	           s, n.ndigits);
 	  FREE (s, n.ndigits + 1);
 	}
 
-      if ((!breadthfirst && cnt == count) || (breadthfirst && 1 == breadthfirst_cnt))
+      if ((!breadthfirst && cnt == count) || 
+          (breadthfirst && 1 == breadthfirst_cnt))
 	{
 	  int SomeFactor;
 	  /*  Note, if a factors are found, then n will be adjusted "down" */
@@ -1105,7 +1239,8 @@ BreadthFirstDoAgain:;
 	      fflush (stdout);
 	      if (!deep)
 		{
-		  /* Note, if we are not in deep mode, then there is no need to continue if a factor was found */
+		  /* Note, if we are not in deep mode, then there is no need 
+		     to continue if a factor was found */
   		  factor_is_prime = 1;
   		  mpz_set_ui (f,1);
 		  goto OutputFactorStuff;
@@ -1120,7 +1255,7 @@ BreadthFirstDoAgain:;
 
       mpgocandi_fixup_with_N (&go, &n);
 
-      /* set parameters that mau change from one curve to another */
+      /* set parameters that may change from one curve to another */
       params->method = method; /* may change with resume */
       mpz_set (params->x, x); /* may change with resume */
       /* if sigma is zero, then we use the A value instead */
@@ -1129,6 +1264,46 @@ BreadthFirstDoAgain:;
       mpz_set (params->go, go.Candi.n); /* may change if contains N */
       params->B1done = B1done; /* may change with resume */
       mpz_set (params->B2min, B2min); /* may change with -c */
+      /* Here's an ugly hack to pass B2scale to the library somehow.
+         It gets piggy-backed onto B1done */
+      params->B1done = B1done + floor (B2scale * 128.) / 134217728.; 
+      /* Default, for P-1/P+1 with old stage 2 and ECM, use NTT only 
+         for small input */
+      if (use_ntt == 1 && (method == ECM_ECM || S != ECM_DEFAULT_S)) 
+        params->use_ntt = (mpz_size (n.n) <= NTT_SIZE_THRESHOLD);
+      else 
+        params->use_ntt = (use_ntt != 0);
+
+#ifdef WANT_SHELLCMD
+      /* See if the system is currently idle, if -idlecmd was given */
+      if (idlecmd != NULL)
+        {
+          int r;
+          FILE *fc;
+          fc = popen (idlecmd, "r");
+          if (fc == NULL)
+            {
+              fprintf (stderr, "Error executing idle command: %s\n",
+                       idlecmd);
+              exit (EXIT_FAILURE);
+            }
+          r = pclose (fc);
+          if (r != 0) /* If exit status of idle command is non-zero */
+            {
+              printf ("Idle command returned %d, exiting\n", r);
+              breadthfirst = 0; /* Avoid looping due to goto (ugly, FIXME!) */
+              break;
+            }
+        }
+#endif /* WANT_SHELLCMD */
+
+      if (timestamp)
+        {
+          time_t t;
+          
+          t = time (NULL);
+          printf ("[%.24s]\n", ctime (&t));
+        }
 
       /* now call the ecm library */
       result = ecm_factor (f, n.n, B1, params);
@@ -1140,33 +1315,56 @@ BreadthFirstDoAgain:;
           exit (EXIT_FAILURE);
         }
 
-      if (result == 0)
+      if (result == ECM_NO_FACTOR_FOUND)
 	{
 	  if (trial_factor_found)
 	  {
 	    factor_is_prime = 1;
 	    mpz_set_ui (f, 1);
+	    returncode = ECM_NO_FACTOR_FOUND;
 	    goto OutputFactorStuff;
 	  }
-	}
-      if (result != 0)
-	{
+	} else {
 	  factsfound++;
 	  if (verbose > 0)
 	    printf ("********** Factor found in step %u: ", ABS (result));
           mpz_out_str (stdout, 10, f);
 	  if (verbose > 0)
             printf ("\n");
-	  if (mpz_cmp (f, n.n))
+
+          /* Complain about non-proper factors (1, 0, negative) */
+          if (mpz_cmp_ui (f, 2) < 0)
+            {
+              fprintf (stderr, "Error: factor found is ");
+              mpz_out_str (stderr, 10, f);
+              fprintf (stderr, "\nPlease report internal errors at <%s>.\n",
+                       PACKAGE_BUGREPORT);
+              exit (EXIT_FAILURE);
+            }
+          
+#ifdef WANT_SHELLCMD
+	  if (faccmd != NULL)
 	    {
-              if (mpz_cmp_ui (f, 1) == 0)
-                {
-                  fprintf (stderr, "Error: factor found is 1\n");
-                  /* THIS SHOULD NOT BE HERE!!! */
-                  exit (1);
-                }
-	      /* prints factor found and cofactor on standard error. */
-	      factor_is_prime = smart_probab_prime_p (f, PROBAB_PRIME_TESTS);
+	      FILE *fc;
+	      fc = popen (faccmd, "w");
+	      if (fc != NULL)
+	        {
+	          mpz_t cof;
+	          mpz_init_set (cof, n.n);
+	          mpz_divexact (cof, cof, f);
+	          gmp_fprintf (fc, "%Zd\n", n.n);
+	          gmp_fprintf (fc, "%Zd\n", f);
+	          gmp_fprintf (fc, "%Zd\n", cof);
+	          mpz_clear (cof);
+	          pclose (fc);
+	        }
+	    }
+#endif
+
+	  if (mpz_cmp (f, n.n) != 0)
+	    {
+	      /* prints factor found and cofactor on standard output. */
+	      factor_is_prime = probab_prime_p (f, PROBAB_PRIME_TESTS);
 
               if (verbose >= 1)
                 {
@@ -1179,6 +1377,21 @@ BreadthFirstDoAgain:;
 
 	      mpcandi_t_addfoundfactor (&n, f, 1); /* 1 for display warning if factor does not divide the current candidate */
 
+              if (resumefile != NULL)
+                {
+                  /* If we are resuming from a save file, add factor to the
+                     discovered factors for the current number */
+                  mpz_mul (resume_lastfac, resume_lastfac, f);
+                  resume_wasPrp = n.isPrp;
+                }
+
+              if (factor_is_prime)
+                returncode = (n.isPrp) ? ECM_PRIME_FAC_PRIME_COFAC : 
+		                         ECM_PRIME_FAC_COMP_COFAC;
+              else
+                returncode = (n.isPrp) ? ECM_COMP_FAC_PRIME_COFAC :
+		                         ECM_COMP_FAC_COMP_COFAC;
+
 OutputFactorStuff:;
 	      if (verbose >= 1)
 		{
@@ -1222,13 +1435,14 @@ OutputFactorStuff:;
 	  else
 	    {
 	      if (breadthfirst)
-		/* I know it may not be prp, but setting this will cause all future loops to NOT 
-		   check this candidate again */
+		/* I know it may not be prp, but setting this will cause all 
+		   future loops to NOT check this candidate again */
 		pCandidates[linenum-1].isPrp = 1;
 	      cnt = 0; /* no more curve to perform */
               if (verbose > 0)
                 printf ("Found input number N");
               printf ("\n");
+              returncode = ECM_INPUT_NUMBER_FOUND;
 	    }
 	  fflush (stdout);
 	}
@@ -1248,16 +1462,19 @@ OutputFactorStuff:;
       /* If no factor was found, we consider cofactor composite and write it */
       if (savefile != NULL && !n.isPrp)
         {
-          mpz_mod (x, params->x, n.n); /* Reduce stage 1 residue wrt new
-                                          cofactor, in case a factor was found */
-          write_resumefile_line (savefile, method, B1, sigma, A, x, &n, 
-                                 orig_x0, comment);
+          mpz_mod (x, params->x, n.n); /* Reduce stage 1 residue wrt new co-
+                                          factor, in case a factor was found */
+          /* We write the B1done value to the safe file. This requires that
+             a correct B1done is returned by the factoring functions */
+          write_resumefile_line (savefile, method, params->B1done, sigma, 
+                                 A, x, &n, orig_x0, comment);
         }
 
-      /* Clean up any temp file left over.  At this time, we "assume" that if the user wants
-         to resume the run, then they used -save file.  The temp save was ONLY to help in case
-         of a power outage (or similar) for a long run.  It would allow finishing the current
-         candidate, keeping the existing work done.   Now, we assume we "are" done. */
+      /* Clean up any temp file left over.  At this time, we assume that if 
+         the user wants to resume the run, then they used -save file.  The 
+         temp save was ONLY to help in case of a power outage (or similar) 
+         for a long run.  It would allow finishing the current candidate, 
+         keeping the existing work done.   Now, we assume we are done. */
 #if !defined (DEBUG_AUTO_SAVE)
       kill_temp_resume_file ();
 #endif
@@ -1274,19 +1491,23 @@ OutputFactorStuff:;
     }
 
   /* Allow our "breadthfirst" search to re-run the file again if enough curves have not yet been run */
-  if (breadthfirst == 1)
+  if (breadthfirst == 1 && !exit_asap_value)
     goto BreadthFirstDoAgain;
 
-  /* NOTE finding a factor may have caused the loop to exit, but what is left on screen is the 
-     wrong count of factors (missing the just found factor.  Update the screen to at least specify the 
-     current count */
+  /* NOTE finding a factor may have caused the loop to exit, but what is left 
+     on screen is the wrong count of factors (missing the just found factor.  
+     Update the screen to at least specify the current count */
 
-  if (infilename)	/* note infile "might" be stdin, and don't fclose that! */
+  if (infilename)	/* infile might be stdin, don't fclose that! */
     fclose (infile);
   if (savefile)
     fclose (savefile);
   if (resumefile)
-    fclose (resumefile);
+    {
+      fclose (resumefile);
+      mpz_clear (resume_lastN);
+      mpz_clear (resume_lastfac);
+    }
   if (nCandidates)
     {
       while (nCandidates--)
@@ -1316,6 +1537,12 @@ OutputFactorStuff:;
   tests_memory_end ();
 #endif
 
-  /* exit 0 iff a factor was found for the last input */
-  return result ? 0 : 1;
+  /* exit 0 if a factor was found for the last input, except if we exit due
+     to a signal */
+#ifdef HAVE_SIGNAL
+  if (returncode == 0 && exit_asap_value != 0)
+    returncode = 143;
+#endif
+
+  return returncode;
 }
